import base64
import json
import re
from io import BytesIO
from typing import List, Optional

import cv2
from jsonschema import SchemaError, ValidationError, validate
from openai import OpenAI
from PIL import Image


def _encode_pil_to_base64(image: Image.Image) -> str:
    """Encodes a PIL Image as a base64-encoded JPEG string.

    This function does not include the data URL prefix, only the base64-encoded
    bytes.

    Args:
        image (PIL.Image.Image): The PIL Image to encode.

    Returns:
        str: Base64-encoded JPEG representation of the image.
    """
    with BytesIO() as buffer:
        image.save(buffer, format="JPEG")
        buffer.seek(0)
        return base64.b64encode(buffer.read()).decode("utf-8")


def _extract_frames_from_video(
    video_path: str, target_fps: float = 1.0
) -> List[Image.Image]:
    """Extracts frames from a video at a specified frames-per-second rate.

    Args:
        video_path (str): Path to the video file.
        target_fps (float, optional): Number of frames to extract per second.
            Defaults to 1.0.

    Returns:
        List[PIL.Image.Image]: A list of PIL Image frames extracted from the video.

    Raises:
        FileNotFoundError: If the video cannot be opened or found.
    """
    frames = []
    capture = cv2.VideoCapture(video_path)
    if not capture.isOpened():
        raise FileNotFoundError(f"Could not open video file: {video_path}")

    native_fps = capture.get(cv2.CAP_PROP_FPS) or target_fps
    frame_skip = max(1, int(round(native_fps / target_fps)))

    frame_count = 0
    while True:
        ret, frame = capture.read()
        if not ret:
            break
        if frame_count % frame_skip == 0:
            frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            frames.append(Image.fromarray(frame_rgb))
        frame_count += 1

    capture.release()
    return frames


def _clean_model_output(raw_output_str: str) -> str:
    """Cleans up a model's raw string output.

    This function removes surrounding triple backticks (with optional
    language tags), then strips leading/trailing whitespace.

    Args:
        raw_output_str (str): The raw output string from the model.

    Returns:
        str: The cleaned string without triple backticks.
    """
    cleaned = raw_output_str.strip()
    cleaned = re.sub(r"^```[a-zA-Z]*\s*", "", cleaned)
    cleaned = re.sub(r"```$", "", cleaned)
    return cleaned.strip()


def _validate_and_extract_json(raw_output_str: str, schema_dict: dict) -> str:
    """Parses, validates, and extracts JSON data from the model output.

    If the parsed JSON contains a top-level 'properties' key, that is
    treated as the real data. Otherwise, the entire parsed JSON is considered
    the real data. The function then validates the data against `schema_dict`.

    Args:
        raw_output_str (str): The raw (already cleaned) JSON string from the model.
        schema_dict (dict): The JSON schema for validation.

    Returns:
        str: A JSON string (pretty-printed with indent=2) if valid.

    Raises:
        RuntimeError: If the JSON is invalid or does not match the schema.
    """
    try:
        parsed_data = json.loads(raw_output_str)
    except json.JSONDecodeError as exc:
        raise RuntimeError(
            f"Model output is not valid JSON: {exc}\n\nRaw output:\n{raw_output_str}"
        ) from exc

    real_data = parsed_data.get("properties", parsed_data)

    try:
        validate(instance=real_data, schema=schema_dict)
    except (ValidationError, SchemaError) as exc:
        raise RuntimeError(
            f"Schema validation failed: {exc}\n\nRaw output:\n{raw_output_str}"
        ) from exc

    return json.dumps(real_data, indent=2)


def video2json(
    video_path: str,
    schema_str: str,
    context_str: Optional[str] = None,
    api_key: str = "",
    model: str = "gpt-4o",
    base_url: Optional[str] = None,
    target_fps: float = 1.0,
    validate_json: bool = True,
) -> str:
    """Extracts frames from a video and generates JSON output from a model.

    Builds a prompt referencing a JSON schema (and optional context), sends
    frames and prompt to an OpenAI-compatible chat endpoint, and returns
    either the validated JSON or the raw model output.

    Args:
        video_path (str): Path to the video file.
        schema_str (str): A string containing the JSON schema.
        context_str (str, optional): Additional context to include in the prompt.
            Defaults to None.
        api_key (str, optional): The API key for authentication. Defaults to "".
        model (str, optional): The name of the model to invoke (e.g. 'gpt-4o').
            Defaults to "gpt-4o".
        base_url (str, optional): Override for the OpenAI-compatible API base URL.
            Defaults to None for standard OpenAI.
        target_fps (float, optional): Frames per second to extract from the video.
            Defaults to 1.0.
        validate_json (bool, optional): If True, validates the model's response against
            the schema. Defaults to True.

    Returns:
        str: The final JSON string if `validate_json` is True, otherwise
            the raw model output.

    Raises:
        ValueError: If no API key is provided or if no frames could be extracted.
        FileNotFoundError: If the video file is not found or cannot be opened.
        RuntimeError: If the model call fails, no choices are returned, or the
            schema validation fails.
    """
    if not api_key:
        raise ValueError("API key must be provided.")

    frames = _extract_frames_from_video(video_path, target_fps)
    if not frames:
        raise ValueError(f"No frames extracted from '{video_path}'.")

    prompt_text = (
        "Your task is to analyze the following images (extracted from a video) and "
        "output a single JSON covering all images (do not return one JSON per image).\n\n"
        f"The JSON must strictly follow this schema:\n{schema_str}"
    )
    if context_str:
        prompt_text += f"\n\nContext:\n{context_str}"

    content_list = [{"type": "text", "text": prompt_text}]
    for frame in frames:
        encoded_image = _encode_pil_to_base64(frame)
        content_list.append(
            {
                "type": "image_url",
                "image_url": {"url": f"data:image/jpeg;base64,{encoded_image}"},
            }
        )

    try:
        client = OpenAI(api_key=api_key, base_url=base_url)
        response = client.chat.completions.create(
            model=model, messages=[{"role": "user", "content": content_list}]
        )
    except Exception as exc:
        raise RuntimeError(f"Error during model call: {exc}") from exc

    if not response.choices:
        raise RuntimeError("No choices returned by the model.")

    # Clean the model output
    raw_output_str = _clean_model_output(response.choices[0].message.content)

    if not validate_json:
        return raw_output_str

    # Validate against the schema
    schema_dict = json.loads(schema_str)
    return _validate_and_extract_json(raw_output_str, schema_dict)
